/* Copyright (c) 2004-2006 ActiveState Software Inc.
   See the file LICENSE.txt for licensing information. */

/* JavaScript-side control of the Code Browser. */


//---- globals

var gCodeBrowserMgr = null;
var gCodeBrowserCtrl = null; // commands available when cb focused

var _gCodeBrowser_log = ko.logging.getLogger("cb_js");
//_gCodeBrowser_log.setLevel(ko.logging.LOG_DEBUG);



//---- internal support routines

// Return true iff the two file paths are equal.
function _fequal(a, b) {
// #if PLATFORM == "win"
    return a.toLowerCase() == b.toLowerCase();
// #else
    return a == b;
// #endif
}


//---- the controller for the code browser commands

function _isNodeSelected() {
    return (gCodeBrowserMgr && gCodeBrowserMgr.view.selection.count != 0);
}

function CodeBrowserController() {
}
// The following two lines ensure proper inheritance (see Flanagan, p. 144).
CodeBrowserController.prototype = new Controller();
CodeBrowserController.prototype.constructor = CodeBrowserController;

CodeBrowserController.prototype.destructor = function() {
}

CodeBrowserController.prototype.is_cmd_copy_enabled = _isNodeSelected;
CodeBrowserController.prototype.do_cmd_copy = function() {
    gCodeBrowserMgr.copySelection();
}


function CodeIntelController() {
}

// The following two lines ensure proper inheritance (see Flanagan, p. 144).
CodeIntelController.prototype = new Controller();
CodeIntelController.prototype.constructor = CodeIntelController;

CodeIntelController.prototype.destructor = function() {
}

CodeIntelController.prototype.is_cmd_codeBrowserGoToDefinition_enabled = _isNodeSelected;
CodeIntelController.prototype.do_cmd_codeBrowserGoToDefinition = function() {
    gCodeBrowserMgr._jumpToSymbol();
}


CodeIntelController.prototype.is_cmd_codeBrowserLocate_enabled = function() {
    if (!gCodeIntelActive) return false;
    var view = ko.views.manager.currentView;
    return (view && view.scimoz && view.document &&
            view.isCICitadelStuffEnabled);
}

CodeIntelController.prototype.do_cmd_codeBrowserLocate = function() {
    CodeIntel_LocateCurrentScope();
}


//---- the UI Manager for the Code Browser (koICodeBrowserManager)

function CodeBrowserManager(treeId)
{
    _gCodeBrowser_log.debug("CodeBrowserManager(treeId='"+treeId+"')");
    try {
        this.widgets = new Object();
        this.widgets.sortByButton = document.getElementById("codebrowser-sort-by-button");
        this.widgets.filterTextbox = document.getElementById("codebrowser-filter-textbox");
        this.widgets.tree = document.getElementById("codebrowser-tree");

        // Attach the nsITreeView implementation to the <tree/>.
        //XXX Do we have to delay this until we are visible for the first
        //    time? If so then make the .visible setter a method and change
        //    the call in CodeIntel_UpdateCodeBrowserVisibility().
        this.view = Components.classes["@activestate.com/koCodeBrowserTreeView;1"]
                              .createInstance(Components.interfaces.koICodeBrowserTreeView);
        this.view.setCodeBrowserMgr(this);
        this.boxObject = this.widgets.tree.treeBoxObject;
        this.boxObject.view = this.view;

        // Restore preferences.
        this.view.restorePrefs();

        // Restore sort order.
        var key = null;
        if (gPrefs.hasStringPref("codeintel_sort_codebrowser_by")) {
            key = gPrefs.getStringPref("codeintel_sort_codebrowser_by");
        }
        if (! this.sortBy(key)) {
            var fallback = "file-order";
            _gCodeBrowser_log.warn("invalid Code Browser sort order key from "+
                                 "prefs ('"+key+"'): falling back to '"+
                                 fallback+"'");
            this.sortBy(fallback);
        }

        // Setup controllers.
        this.cbController = new CodeBrowserController();
        this.widgets.tree.controllers.insertControllerAt(0, this.cbController);
        this.ciController = new CodeIntelController();
        window.controllers.insertControllerAt(0, this.ciController);

        // Setup event listeners.
        var this_ = this;
        this._on_current_view_language_changed = function(event) {
            var view = event.originalTarget;
            this_.view.change_buf_lang(view.document.ciBuf);
            window.setTimeout("window.updateCommands('codebrowser');", 0);            
        }
        this._on_view_document_attached = function(event) {
            var view = event.originalTarget;
            if (view.getAttribute("type") != "editor") {
                return;
            }
            this_.view.add_buf(view.document.ciBuf);
            window.setTimeout("window.updateCommands('codebrowser');", 0);            
        }
        this._on_view_document_detaching = function(event) {
            var view = event.originalTarget;
            if (view.getAttribute("type") != "editor") {
                return;
            }
            this_.view.remove_buf(view.document.ciBuf);
            window.setTimeout("window.updateCommands('codebrowser');", 0);            
        }
        window.addEventListener('current_view_language_changed',
            this._on_current_view_language_changed, false);
        window.addEventListener('view_document_attached', this._on_view_document_attached, false);
        window.addEventListener('view_document_detaching', this._on_view_document_detaching, false);        

        this.visible = null; // is the "Code" tab visible?
        this._mouseMoveTimer = null;
        
        // When the code browser starts up there may already be open files
        // in the current Komodo window.
        this._showTab();
        window.setTimeout(function (this_) {
            this_._addCurrentDocuments();
        }, 2000, this);
    } catch(ex) {
        _gCodeBrowser_log.exception(ex);
    }
}
CodeBrowserManager.prototype.constructor = CodeBrowserManager;

CodeBrowserManager.prototype.QueryInterface = function(iid) {
    if (!iid.equals(Components.interfaces.koICodeBrowserManager) &&
        !iid.equals(Components.interfaces.nsISupports)) {
        throw Components.results.NS_ERROR_NO_INTERFACE;
    }
    return this;
}

CodeBrowserManager.prototype.finalize = function() {
    _gCodeBrowser_log.debug("CodeBrowserManager.finalize()");
    try {
        this.view.savePrefs();
        this.view.finalize();

        gPrefs.setStringPref("codeintel_sort_codebrowser_by",
                             this.widgets.sortByButton.getAttribute("sortBy"));
        this._hideTab();

        window.removeEventListener("view_document_attached",
            this._on_view_document_attached, false);
        window.removeEventListener("view_document_detaching",
            this._on_view_document_detaching, false);
        window.removeEventListener("current_view_language_changed",
            this._on_current_view_language_changed, false);        

        this.widgets.tree.controllers.removeController(this.cbController);
        window.controllers.removeController(this.ciController);
    } catch(ex) {
        _gCodeBrowser_log.exception(ex);
    }
}

CodeBrowserManager.prototype._addCurrentDocuments = function() {
    _gCodeBrowser_log.debug("CodeBrowserManager._addCurrentDocuments()");

    var i, doc, buf = 42;
    var editorViews = ko.views.manager.topView.getViewsByType(true, "editor");
    for (i = 0; i < editorViews.length; i++) {
        doc = editorViews[i].document;
        buf = doc.ciBuf;
        if (buf && gCodeIntelSvc.is_citadel_lang(buf.lang)) {
            this.view.add_buf(buf);
            gCodeIntelSvc.scan_document(doc, 0, true);
        }
        
    }
}

// Show the Code Browser tab.
CodeBrowserManager.prototype._showTab = function() {
    _gCodeBrowser_log.debug("CodeBrowserManager._showTab()");

    var tab = document.getElementById("codebrowser_tab");
    if (tab.hasAttribute("collapsed"))
        tab.removeAttribute("collapsed");
    // <tabbox> Ctrl+Tab handling uses 'hidden' attribute.
    if (tab.hasAttribute("hidden"))
        tab.removeAttribute("hidden");
}

// Hide the Code Browser tab.
CodeBrowserManager.prototype._hideTab = function() {
    _gCodeBrowser_log.debug("CodeBrowserManager._hideTab()");

    var tabbox = document.getElementById("leftTabBox");
    var tab = document.getElementById("codebrowser_tab");
    if (tabbox.selectedTab == tab) {
        var alternate = tab.previousSibling;
        if (!alternate)
            alternate = tab.nextSibling;
        if (alternate)
            tabbox.selectedTab = alternate;
    }
    tab.setAttribute("collapsed", "true");
    // <tabbox> Ctrl+Tab handling uses 'hidden' attribute.
    tab.setAttribute("hidden", "true");
}

CodeBrowserManager.prototype.onSelect = function(event) {
    try {
        gCodeBrowserMgr.updateCommands();
    } catch(ex) {
        _gCodeBrowser_log.exception(ex);
    }
}

CodeBrowserManager.prototype.onFilterKeypress = function(event) {
    try {
        if (event.keyCode == event.DOM_VK_TAB && !event.ctrlKey) {
            event.cancelBubble = true;
            event.stopPropagation();
            event.preventDefault();
            gCodeBrowserMgr.widgets.tree.focus();
            return;
        } else if (event.keyCode == event.DOM_VK_ESCAPE) {
            if (this.widgets.filterTextbox.value != '') {
                this.widgets.filterTextbox.value = '';
                gCodeBrowserMgr.updateFilter();
                event.cancelBubble = true;
                event.stopPropagation();
                event.preventDefault();
            }
            return;
        }
    } catch (e) {
        _gCodeBrowser_log.exception(e);
    }
}

CodeBrowserManager.prototype.updateCommands = function() {
    // update commands when the selection changes.
    ko.commands.updateCommand('cmd_copy', null, gCodeBrowserCtrl);
    ko.commands.updateCommand('cmd_codeBrowserGoToDefinition', null, gCodeIntelCtrl);
}

CodeBrowserManager.prototype.focus = function() {
    _gCodeBrowser_log.debug("CodeBrowserManager.focus()");
    try {
        // Pass the focus onto the <tree/>.
        // It is important to do this because this focus handling is used to
        // get ko.uilayout.toggleTab() to work properly.
        this.widgets.tree.focus();
    } catch(ex) {
        _gCodeBrowser_log.exception(ex);
    }
}

CodeBrowserManager.prototype._jumpToSymbol = function() {
    _gCodeBrowser_log.debug("CodeBrowserManager._jumpToSymbol()");
    try {
        var path = this.view.currentNodeFilePath;
        if (path == null) {
            return;
        }
        var line = this.view.currentNodeLine;

        if (path.slice(0, 9) == "<Unsaved>") {
            //XXX Komodo's view system does not currently provide a simple
            //    method to open or get a handle on the view for a given
            //    untitled document. We therefore have to hack it ourselves.
            path = path.slice(10);
            var views = ko.views.manager.topView.getViews(true);
            var v;
            for (var i = 0; i < views.length; i++) {
                v = views[i];
                if (v.getAttribute("type") == "editor"
                    && v.document.isUntitled
                    && _fequal(v.document.baseName, path))
                {
                    v.makeCurrent();
                    if (line) v.currentLine = line;
                    break;
                }
            }
        } else {
            var uri = ko.uriparse.pathToURI(path);
            if (line) {
                ko.views.manager.doFileOpenAtLineAsync(uri, line, 'editor',
                                                       null, -1,
                                                       function(view) {
                    // Bug 32471: scroll line to a consistent position: 30% down
                    // from top. (This is a SciMoz recipe to scroll a given line
                    // to 30% down from the top of the viewport, properly
                    // accounting for folding and line wrapping.)
            
                    // XXX : Move this into the doFileOpenAtLine code.
                    var scimoz = view.scimoz;
                    scimoz.ensureVisible(line);
                    var want2BNLinesFromTop = Math.round(scimoz.linesOnScreen*0.30);
                    var wantTopLine2B = Math.max(0,
                        scimoz.visibleFromDocLine(line) - want2BNLinesFromTop);
                    var scrollBy = wantTopLine2B - scimoz.firstVisibleLine;
                    //dump("- Want line "+line+" (considering folding: "
                    //     +scimoz.visibleFromDocLine(line)+") to be "
                    //     +want2BNLinesFromTop+" lines (30%) from top.\n"
                    //     +"  I.e. want top line to be "+
                    //     scimoz.docLineFromVisible(wantTopLine2B)
                    //     +" (considering folding: "+wantTopLine2B+"),\n"
                    //     +"  therefore need to scroll "+scrollBy+" lines.\n");
                    scimoz.lineScroll(0, scrollBy);
                });
            } else {
                ko.views.manager.doFileOpenAsync(uri);
            }
        }
    } catch(ex) {
        _gCodeBrowser_log.exception(ex);
    }
}


CodeBrowserManager.prototype.onTreeKeyPress = function(event) {
    _gCodeBrowser_log.debug("CodeBrowserManager.onTreeKeyPress(event)");
    try {
        if (event.keyCode == event.DOM_VK_TAB && !event.ctrlKey)
        {
            event.cancelBubble = true;
            event.stopPropagation();
            event.preventDefault();
            window.setTimeout("gCodeBrowserMgr.focusFilterBox();", 0);
            return;
        }
        if (event.keyCode == event.DOM_VK_ENTER
            || event.keyCode == event.DOM_VK_RETURN)
        {
            event.cancelBubble = true;
            event.stopPropagation();
            event.preventDefault();
            gCodeBrowserMgr._jumpToSymbol();
        }
    } catch(ex) {
        _gCodeBrowser_log.exception(ex);
    }
}

CodeBrowserManager.prototype.focusFilterBox = function() {
    this.widgets.filterTextbox.focus();
}


// Create instance of component given contractId and iid (as string).
CodeBrowserManager.prototype._createInstance = function(contractId, iidName) {
        var iid = Components.interfaces[iidName];
        return Components.classes[contractId].createInstance(iid);
}
// Get service given contractId and iid (as string).
CodeBrowserManager.prototype._getService = function(contractId, iidName) {
        var iid = Components.interfaces[iidName];
        return Components.classes[contractId].getService(iid);
}

CodeBrowserManager.prototype._copyToClipboard = function(text) {
    // Copy link/image url to clipboard.
    // Get clipboard.
    var clipboard = this._getService("@mozilla.org/widget/clipboard;1",
                                     "nsIClipboard");

    // Create tranferable that will transfer the text.
    var transferable = this._createInstance("@mozilla.org/widget/transferable;1",
                                            "nsITransferable");

    if (clipboard && transferable) {
        transferable.addDataFlavor("text/unicode");
        // Create wrapper for text.
        var data = this._createInstance("@mozilla.org/supports-string;1",
                                        "nsISupportsString");
        if (data) {
            data.data = text;
            transferable.setTransferData("text/unicode", data, text.length * 2);
            // Put on clipboard.
            clipboard.setData(transferable, null,
                              Components.interfaces.nsIClipboard.kGlobalClipboard);
        }
    }
}

CodeBrowserManager.prototype.copySelection = function() {
    try {
        var currentIndex = this.view.selection.currentIndex;
        if (currentIndex == -1) {
            return;
        }
        var column;
// #if MOZILLA_VERSION > "1.7"
        column = this.boxObject.columns.getNamedColumn("codebrowser-tree-node");
// #else
        column = "codebrowser-tree-node";
// #endif
        var text = this.view.getCellText(currentIndex, column);
        this._copyToClipboard(text);
    } catch (e) {
        log.exception(e);
    }
}


CodeBrowserManager.prototype.onDblClick = function(event) {
    _gCodeBrowser_log.debug("CodeBrowserManager.onDblClick(event)");
    try {
        this._jumpToSymbol();
    } catch(ex) {
        _gCodeBrowser_log.exception(ex);
    }
}

// Let the Code Browser tree view know if the mouse is over a specific
// node.  The view can use this info to show more information
// for that node.
CodeBrowserManager.prototype._handleMouseDwell = function(x, y) {
    try {
        var rowObj = new Object();
        var colObj = new Object();
        var childElementObj = new Object();
        this.boxObject.getCellAt(x, y, rowObj, colObj, childElementObj);
        var row = rowObj.value;
        var col = colObj.value;
        var childElement = childElementObj.value;

        // assert col == "codebrowser-tree-node"
        gCodeBrowserMgr.view.mouseOverNode(row, childElement);

        this._mouseMoveTimer.free();
        this._mouseMoveTimer = null;
    } catch(ex) {
        _gCodeBrowser_log.exception(ex);
    }
}
CodeBrowserManager.prototype.onMouseMove = function(event) {
    _gCodeBrowser_log.debug("CodeBrowserManager.onMouseMove(event)");
    try {
        if (event.originalTarget.localName == "treechildren") {
            if (this._mouseMoveTimer) {
                this._mouseMoveTimer.stopTimeout();
                this._mouseMoveTimer.free();
                this._mouseMoveTimer = null;
            }
            this._mouseMoveTimer = new objectTimer(this, this._handleMouseDwell,
                                                   [event.clientX, event.clientY]);
            this._mouseMoveTimer.startTimeout(100);
        }
    } catch(ex) {
        _gCodeBrowser_log.exception(ex);
    }
}
CodeBrowserManager.prototype.onMouseOut = function(event) {
    _gCodeBrowser_log.debug("CodeBrowserManager.onMouseOut(event)");
    try {
        window.setTimeout("gCodeBrowserMgr.view.mouseOverNode(-1, '');", 50);
    } catch(ex) {
        _gCodeBrowser_log.exception(ex);
    }
}

// Set a new sort order for the code browser.
// Returns true iff successful.
CodeBrowserManager.prototype.sortBy = function(key) {
    _gCodeBrowser_log.debug("CodeBrowserManager.sortBy(key='"+key+"')");
    try {
        this._validSortByKeys = ["file-order",
                                 "name"];
        var valid = false;
        for (var i = 0; i < this._validSortByKeys.length; ++i) {
            if (key == this._validSortByKeys[i]) {
                valid = true;
                break;
            }
        }

        if (valid) {
            this.view.sortBy(key);
            this.widgets.sortByButton.setAttribute("sortBy", key);
        }
        return valid;
    } catch(ex) {
        _gCodeBrowser_log.exception(ex);
        return false;
    }
}


CodeBrowserManager.prototype.haveFilesInWS = function(value) {
    _gCodeBrowser_log.debug("CodeBrowserManager.haveFilesInWS(value='"+value+"')");
    try {
        if (value) {
            if (this.widgets.sortByButton.hasAttribute("disabled"))
                this.widgets.sortByButton.removeAttribute("disabled");
            if (this.widgets.filterTextbox.hasAttribute("disabled"))
                this.widgets.filterTextbox.removeAttribute("disabled");
            // Ensure that there is a selection in the tree.
            if (!this.view.selection.count) { // there is no selection
                this.view.selection.select(0);
            }
        } else {
            this.widgets.sortByButton.setAttribute("disabled", "true");
            this.widgets.filterTextbox.setAttribute("disabled", "true");
        }
    } catch(ex) {
        _gCodeBrowser_log.exception(ex);
    }
}


CodeBrowserManager.prototype.updateFilter = function() {
    _gCodeBrowser_log.debug("CodeBrowserManager.updateFilter()");
    try {
        var textbox = this.widgets.filterTextbox;
        var filterPattern = textbox.value;
        try {
            this.view.setFilter(filterPattern);
            if (textbox.hasAttribute("error")) {
                textbox.removeAttribute("error");
                textbox.setAttribute("tooltiptext",
                    textbox.getAttribute("basetooltiptext"));
            }
        } catch (ex) { // Means an invalid regex pattern.
            textbox.setAttribute("error", "true");
            textbox.setAttribute("tooltiptext",
                textbox.getAttribute("basetooltiptext") + " -- error in pattern: " + ex);
        }
        if (this.view.rowCount) {
            this.view.selection.select(0);
        }
        this.updateCommands();
    } catch(ex) {
        _gCodeBrowser_log.exception(ex);
    }
}

// Locate the appropriate node for the current scope in the Code Browser.
//XXX old way (USE_CITREE_EVAL=False): function(scimoz, position)
CodeBrowserManager.prototype.locateScope = function(buf, pos) {
    _gCodeBrowser_log.debug("CodeBrowserManager.locateScope(buf, pos)");
    try {
        var located = this.view.locateScope(buf, pos);
        if (located) {
            var filter = document.getElementById("codebrowser-filter-textbox");
            filter.value = "";
            ko.uilayout.ensureTabShown("codebrowser_tab", true);
        }
        return located;
    } catch(ex) {
        _gCodeBrowser_log.exception(ex);
        return 0;
    }
}



